""" A kernel transformation that collects data until the vector size is reached """

from dune.perftool.loopy.vcl import VCLLoad, VCLStore
from dune.perftool.tools import get_pymbolic_basename

from loopy.symbolic import pw_aff_to_expr

from pymbolic.mapper.dependency import DependencyMapper
from pymbolic.mapper.substitutor import substitute

import pymbolic.primitives as prim
import loopy as lp
import numpy as np


def load_instruction(vec, basepointer, offset=0, **kwargs):
    # TODO: What I really want to pass in here is a pointer type
    pointer = basepointer
    if offset:
        pointer = prim.Sum((pointer, offset))
    return lp.CallInstruction((),
                              prim.Call(VCLLoad(vec), (pointer,)),
                              tags=kwargs.get('tags', frozenset()).union(frozenset(["vcl_load"])),
                              **kwargs
                              )


def store_instruction(vec, basepointer, offset=0, **kwargs):
    # TODO: What I really want to pass in here is a pointer type
    pointer = basepointer
    if offset:
        pointer = prim.Sum((pointer, offset))
    return lp.CallInstruction((),
                              prim.Call(VCLStore(vec), (pointer,)),
                              tags=kwargs.get('tags', frozenset()).union(frozenset(["vcl_store"])),
                              **kwargs
                              )


def collect_vector_data(knl, insns, inames, vec_size=4):
    # TODO: In theory this vec_size should be deduced from the types
    # that we find in the actual instructions

    # Process inames input -> frozenset
    if isinstance(inames, str):
        inames = inames.split(",")
    inames = frozenset(inames)

    # Filter instructions into the groups that we need for this transformation:
    # 1. Those that should be vectorized by this transformation -> insns
    # 2. Writers of quantities that depend on the inames and that 1) need -> write_insns
    # 3. "Outside" instructions, that the instructions from 1)+2) depend on -> dep_insns
    # 4. All other instructions (needed to correctly reconstruct) -> other_insns
    #
    # All these should be lists of instruction instances.

    # 1. Those that should be vectorized by this transformation -> insns
    #
    # Instructions may be passed in as either:
    # * A match expression object
    # * A comma separated string of IDs
    # * An iterable of IDs
    if isinstance(insns, lp.match.MatchExpressionBase):
        insns = lp.find_instructions(knl, insns)
    else:
        if isinstance(insns, str):
            insns = [i.strip() for i in insns.split(",")]
        insns = [knl.id_to_insn[i] for i in insns]

    # 2. Writers of quantities that depend on the inames and that 1) need -> write_insns
    #
    # To determine these, we need to
    # * Find all written quantities in the instructions from 1)
    # * Find the instructions that write to these quantities
    # * Filter only those that depend on the given inames
    quantities = {}
    depmapper = DependencyMapper()
    for insn in insns:
        for expr in depmapper(insn.expression):
            basename = get_pymbolic_basename(expr)
            quantities.setdefault(basename, frozenset())
            quantities[basename] = quantities[basename].union(frozenset([expr]))

    write_match = lp.match.Or(tuple(lp.match.Writes(q) for q in quantities))
    iname_match = lp.match.And(tuple(lp.match.Iname(i) for i in inames))
    match = lp.match.And((write_match, iname_match))
    write_insns = lp.find_instructions(knl, match)

    # 3. "Outside" instructions, that the instructions from 1) depend on -> dep_insns
    deps = frozenset([]).union(*(insn.depends_on for insn in insns + write_insns))
    deps = deps - frozenset([i.id for i in insns + write_insns])
    dep_insns = lp.find_instructions(knl, lp.match.Or(tuple(lp.match.Id(i) for i in deps)))

    # 4. All other instructions (needed to correctly reconstruct) -> other_insns
    used_ids = frozenset([i.id for i in insns + write_insns + dep_insns])
    other_insns = lp.find_instructions(knl, lp.match.Not(lp.match.Or(tuple(lp.match.Id(i) for i in used_ids))))

    #
    # Assert some assumptions on the instructions
    #

    # An instruction occurs in but one of these groups:
    assert len(set(insns + write_insns + dep_insns + other_insns)) == len(insns + write_insns + dep_insns + other_insns)

    # Analyse the inames of the given instructions and identify inames
    # that they all have in common. Those inames will also be iname dependencies
    # of inserted instruction.
    common_inames = frozenset([]).union(*(insn.within_inames for insn in insns)) - inames

    # Have a list gathering all *new* instructions
    new_insns = []
    temporaries = knl.temporary_variables
    keep_rules = []

    # Insert a flat consecutive counter 'total_index'
    temporaries['total_index'] = lp.TemporaryVariable('total_index',  # name
                                                      dtype=np.int32,
                                                      scope=lp.temp_var_scope.LOCAL,
                                                      )
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   id="assign_total_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   prim.Sum((prim.Variable("total_index"), 1)),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset([i.id for i in write_insns]).union(frozenset({"assign_total_index"})),
                                   depends_on_is_final=True,
                                   id="update_total_index",
                                   ))

    # Insert a rotating index, that counts 0 , .. , vecsize - 1
    temporaries['rotate_index'] = lp.TemporaryVariable('rotate_index',  # name
                                                       dtype=np.int32,
                                                       scope=lp.temp_var_scope.LOCAL,
                                                       )
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   id="assign_rotate_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   prim.Remainder(prim.Sum((prim.Variable("rotate_index"), 1)), vec_size),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset([i.id for i in write_insns]).union(frozenset({"assign_rotate_index"})),
                                   depends_on_is_final=True,
                                   id="update_rotate_index",
                                   ))

    # Pre-evaluate all the needed quantities
    replacemap_arr = {}
    replacemap_vec = {}
    for quantity, quantity_exprs in quantities.items():
        # TODO for now I only consider the case where an array occurs but once!
        assert len(quantity_exprs) == 1
        quantity_expr, = quantity_exprs

        # Determine the shape of the

        arrname = quantity + '_buffered_arr'
        temporaries[arrname] = lp.TemporaryVariable(arrname,
                                                    dtype=np.float64,
                                                    shape=(vec_size,),
                                                    dim_tags="c",
                                                    base_storage=quantity + '_base_storage',
                                                    scope=lp.temp_var_scope.LOCAL,
                                                    )

        vecname = quantity + '_buffered_vec'
        temporaries[vecname] = lp.TemporaryVariable(vecname,
                                                    dtype=np.float64,
                                                    shape=(vec_size,),
                                                    dim_tags="vec",
                                                    base_storage=quantity + '_base_storage',
                                                    scope=lp.temp_var_scope.LOCAL,
                                                    )

        replacemap_arr[quantity] = prim.Subscript(prim.Variable(arrname), (prim.Variable('rotate_index'),))
        replacemap_vec[quantity_expr] = prim.Variable(vecname)

    for insn in write_insns:
        if isinstance(insn, lp.Assignment):
            new_insns.append(insn.copy(assignee=replacemap_arr[get_pymbolic_basename(insn.assignee)],
                                       )
                             )
        if isinstance(insn, lp.CInstruction):
            # TODO: What do we do about CInstructions?
            # Example: detjac = ...
            new_insns.append(insn)

    # Determine the condition for the continue statement
    upper_bound = prim.Product(tuple(pw_aff_to_expr(knl.get_iname_bounds(i).size) for i in inames))
    total_check = prim.Comparison(prim.Variable("total_index"), "<", upper_bound)
    rotate_check = prim.Comparison(prim.Variable("rotate_index"), "!=", 0)
    check = prim.LogicalAnd((rotate_check, total_check))

    # Insert the 'continue' statement
    new_insns.append(lp.CInstruction((),  # iname exprs that the code needs access to
                                     "continue;",  # the code
                                     predicates=frozenset({check}),
                                     depends_on=frozenset({"update_rotate_index", "update_total_index"}).union(frozenset([i.id for i in write_insns])),
                                     depends_on_is_final=True,
                                     within_inames=common_inames.union(inames),
                                     within_inames_is_final=True,
                                     id="continue_stmt",
                                     ))

    for insn in insns:
        # Turn the assignee array into a base storage array
        basename = get_pymbolic_basename(insn.assignee)
        assert basename in temporaries
        if temporaries[basename].base_storage is None:
            temporaries[basename] = temporaries[basename].copy(base_storage="{}_base".format(basename))

        # Load the data for this instruction
        load_instructions = []
        for expr in depmapper(insn.expression):
            name = get_pymbolic_basename(expr)
            new_insns.append(load_instruction(replacemap_vec[expr],
                                              replacemap_vec[expr],
                                              depends_on=frozenset({"continue_stmt"}),
                                              within_inames=inames,
                                              within_inames_is_final=True,
                                              id="{}_load_instr".format(name)
                                              )
                             )
            load_instructions.append("{}_load_instr".format(name))

        # TODO do something about the assignee!
        name = "{}_result_buffer".format(insn.id)
        temporaries[name] = lp.TemporaryVariable(name,
                                                 dtype=np.float64,
                                                 shape=(vec_size,),
                                                 dim_tags="vec",
                                                 base_storage="{}_base".format(basename),
                                                 scope=lp.temp_var_scope.LOCAL,
                                                 )
        new_insns.append(insn.copy(assignee=prim.Variable(name),
                                   expression=substitute(insn.expression, variable_assignments=replacemap_vec),
                                   depends_on=insn.depends_on.union(frozenset(["continue_stmt"])).union(frozenset(load_instructions))
                                   )
                         )

        new_insns.append(store_instruction(prim.Variable(name),
                                           prim.Variable(basename),
                                           offset=prim.Sum((prim.Variable("total_index") // vec_size, -1)),
                                           depends_on=frozenset({insn.id}),
                                           within_inames=inames,
                                           within_inames_is_final=True
                                           )
                         )

        keep_rules.append('temp_to_write({})'.format(basename))
        keep_rules.append('read_no_write({})'.format(basename))

    # Return a kernel
    return knl.copy(instructions=dep_insns + other_insns + new_insns,
                    temporary_variables=temporaries,
                    silenced_warnings=knl.silenced_warnings + keep_rules
                    )
